@@ -1,7 +1,7 @@
 /*
- *  This file is part of Poedit (http://poedit.net)
+ *  This file is part of Poedit (https://poedit.net)
  *
- *  Copyright (C) 2013-2015 Vaclav Slavik
+ *  Copyright (C) 2013-2016 Vaclav Slavik
  *
  *  Permission is hereby granted, free of charge, to any person obtaining a
  *  copy of this software and associated documentation files (the "Software"),
@@ -27,6 +27,7 @@
 
 #include "catalog.h"
 #include "errors.h"
+#include "utility.h"
 
 #include <wx/stdpaths.h>
 #include <wx/utils.h>
@@ -64,9 +65,13 @@
 #include <StringReader.h>
 #include <TokenStream.h>
 #include <TermAttribute.h>
+#include <PositionIncrementAttribute.h>
 
 using namespace Lucene;
 
+namespace
+{
+
 #define CATCH_AND_RETHROW_EXCEPTION                                 \
     catch (LuceneException& e)                                      \
     {                                                               \
@@ -78,6 +83,106 @@ using namespace Lucene;
         throw Exception(e.what());                                  \
     }
 
+
+// Manages IndexReader and Searcher instances in multi-threaded environment.
+// Curiously, Lucene uses shared_ptr-based refcounting *and* explicit one as
+// well, with a crucial part not well protected.
+//
+// See https://issues.apache.org/jira/browse/LUCENE-3567 for the exact issue
+// encountered by Poedit's use as well. For an explanation of the manager
+// class, see
+// http://blog.mikemccandless.com/2011/09/lucenes-searchermanager-simplifies.html
+// http://blog.mikemccandless.com/2011/11/near-real-time-readers-with-lucenes.html
+class SearcherManager
+{
+public:
+    SearcherManager(IndexWriterPtr writer)
+    {
+        m_reader = writer->getReader();
+        m_searcher = newLucene<IndexSearcher>(m_reader);
+    }
+
+    ~SearcherManager()
+    {
+        m_searcher.reset();
+        m_reader->decRef();
+    }
+
+    // Safe, properly ref-counting (in Lucene way, not just shared_ptr) holder.
+    template<typename T>
+    class SafeRef
+    {
+    public:
+        typedef boost::shared_ptr<T> TPtr;
+
+        SafeRef(SafeRef&& other) : m_mng(other.m_mng) { std::swap(m_ptr, other.m_ptr); }
+        ~SafeRef() { if (m_ptr) m_mng.DecRef(m_ptr); }
+
+        TPtr ptr() { return m_ptr; }
+        T* operator->() const { return m_ptr.get(); }
+
+        SafeRef(const SafeRef&) = delete;
+        SafeRef& operator=(const SafeRef&) = delete;
+
+    private:
+        friend class SearcherManager;
+        explicit SafeRef(SearcherManager& m, TPtr ptr) : m_mng(m), m_ptr(ptr) {}
+
+        SearcherManager& m_mng;
+        boost::shared_ptr<T> m_ptr;
+    };
+
+    SafeRef<IndexReader> Reader()
+    {
+        std::lock_guard<std::mutex> guard(m_mutex);
+        ReloadReaderIfNeeded();
+        m_reader->incRef();
+        return SafeRef<IndexReader>(*this, m_reader);
+    }
+
+    SafeRef<IndexSearcher> Searcher()
+    {
+        std::lock_guard<std::mutex> guard(m_mutex);
+        ReloadReaderIfNeeded();
+        m_searcher->getIndexReader()->incRef();
+        return SafeRef<IndexSearcher>(*this, m_searcher);
+    }
+
+private:
+    void ReloadReaderIfNeeded()
+    {
+        // contract: m_mutex is locked when this function is called
+        if (m_reader->isCurrent())
+            return; // nothing to do
+
+        auto newReader = m_reader->reopen();
+        auto newSearcher = newLucene<IndexSearcher>(newReader);
+
+        m_reader->decRef();
+
+        m_reader = newReader;
+        m_searcher = newSearcher;
+    }
+
+    void DecRef(IndexReaderPtr& r)
+    {
+        std::lock_guard<std::mutex> guard(m_mutex);
+        r->decRef();
+    }
+    void DecRef(IndexSearcherPtr& s)
+    {
+        std::lock_guard<std::mutex> guard(m_mutex);
+        s->getIndexReader()->decRef();
+    }
+
+    IndexReaderPtr   m_reader;
+    IndexSearcherPtr m_searcher;
+    std::mutex       m_mutex;
+};
+
+
+} // anonymous namespace
+
 // ----------------------------------------------------------------
 // TranslationMemoryImpl
 // ----------------------------------------------------------------
@@ -91,28 +196,32 @@ public:
     typedef MMapDirectory DirectoryType;
 #endif
 
-    TranslationMemoryImpl()
-        : m_dir(GetDatabaseDir()),
-          m_analyzer(newLucene<StandardAnalyzer>(LuceneVersion::LUCENE_CURRENT))
-    {}
+    TranslationMemoryImpl() { Init(); }
 
-    SuggestionsList Search(const Language& lang,
+    ~TranslationMemoryImpl()
+    {
+        m_mng.reset();
+        m_writer->close();
+    }
+
+    SuggestionsList Search(const Language& srclang, const Language& lang,
                            const std::wstring& source);
 
-    std::shared_ptr<TranslationMemory::Writer> CreateWriter();
+    std::shared_ptr<TranslationMemory::Writer> GetWriter() { return m_writerAPI; }
 
     void GetStats(long& numDocs, long& fileSize);
 
 private:
+    void Init();
+
     static std::wstring GetDatabaseDir();
-    IndexReaderPtr Reader();
-    SearcherPtr Searcher();
 
 private:
-    std::wstring     m_dir;
     AnalyzerPtr      m_analyzer;
-    IndexReaderPtr   m_reader;
-    std::mutex       m_readerMutex;
+    IndexWriterPtr   m_writer;
+    std::shared_ptr<SearcherManager> m_mng;
+
+    std::shared_ptr<TranslationMemory::Writer> m_writerAPI;
 };
 
 
@@ -137,17 +246,6 @@ std::wstring TranslationMemoryImpl::GetDatabaseDir()
 }
 
 
-IndexReaderPtr TranslationMemoryImpl::Reader()
-{
-    std::lock_guard<std::mutex> guard(m_readerMutex);
-    if ( m_reader )
-        m_reader = m_reader->reopen();
-    else
-        m_reader = IndexReader::open(newLucene<DirectoryType>(m_dir), true);
-    return m_reader;
-}
-
-
 namespace
 {
 
@@ -168,9 +266,28 @@ bool ContainsResult(const SuggestionsList& all, const std::wstring& r)
     return found != all.end();
 }
 
+// Return translation (or source) text field.
+//
+// Older versions of Poedit used to store C-like escaped text (e.g. "\n" instead
+// of newline), but starting with 1.8, the "true" form of the text is stored.
+// To preserve compatibility with older data, a version field is stored with
+// TM documents and this function decides whether to decode escapes or not.
+//
+// TODO: remove this a few years down the road.
+std::wstring get_text_field(DocumentPtr doc, const std::wstring& field)
+{
+    auto version = doc->get(L"v");
+    auto value = doc->get(field);
+    if (version.empty()) // pre-1.8 data
+        return UnescapeCString(value);
+    else
+        return value;
+}
+
+
 template<typename T>
 void PerformSearchWithBlock(IndexSearcherPtr searcher,
-                            QueryPtr lang,
+                            QueryPtr srclang, QueryPtr lang,
                             const std::wstring& exactSourceText,
                             QueryPtr query,
                             double scoreThreshold,
@@ -178,6 +295,7 @@ void PerformSearchWithBlock(IndexSearcherPtr searcher,
                             T callback)
 {
     auto fullQuery = newLucene<BooleanQuery>();
+    fullQuery->add(srclang, BooleanClause::MUST);
     fullQuery->add(lang, BooleanClause::MUST);
     fullQuery->add(query, BooleanClause::MUST);
 
@@ -191,7 +309,7 @@ void PerformSearchWithBlock(IndexSearcherPtr searcher,
             continue;
 
         auto doc = searcher->doc(scoreDoc->doc);
-        auto src = doc->get(L"source");
+        auto src = get_text_field(doc, L"source");
         if (src == exactSourceText)
         {
             score = 1.0;
@@ -217,7 +335,7 @@ void PerformSearchWithBlock(IndexSearcherPtr searcher,
 }
 
 void PerformSearch(IndexSearcherPtr searcher,
-                   QueryPtr lang,
+                   QueryPtr srclang, QueryPtr lang,
                    const std::wstring& exactSourceText,
                    QueryPtr query,
                    SuggestionsList& results,
@@ -226,11 +344,11 @@ void PerformSearch(IndexSearcherPtr searcher,
 {
     PerformSearchWithBlock
     (
-        searcher, lang, exactSourceText, query,
+        searcher, srclang, lang, exactSourceText, query,
         scoreThreshold, scoreScaling,
         [&results](DocumentPtr doc, double score)
         {
-            auto t = doc->get(L"trans");
+            auto t = get_text_field(doc, L"trans");
             if (!ContainsResult(results, t))
             {
                 time_t ts = DateField::stringToTime(doc->get(L"created"));
@@ -245,11 +363,15 @@ void PerformSearch(IndexSearcherPtr searcher,
 
 } // anonymous namespace
 
-SuggestionsList TranslationMemoryImpl::Search(const Language& lang,
+SuggestionsList TranslationMemoryImpl::Search(const Language& srclang,
+                                              const Language& lang,
                                               const std::wstring& source)
 {
     try
     {
+        // TODO: query by srclang too!
+        auto srclangQ = newLucene<TermQuery>(newLucene<Term>(L"srclang", srclang.WCode()));
+
         const Lucene::String fullLang = lang.WCode();
         const Lucene::String shortLang = StringUtils::toUnicode(lang.Lang());
 
@@ -278,26 +400,28 @@ SuggestionsList TranslationMemoryImpl::Search(const Language& lang,
 
         auto stream = m_analyzer->tokenStream(sourceField, newLucene<StringReader>(source));
         int sourceTokensCount = 0;
+        int sourceTokenPosition = -1;
         while (stream->incrementToken())
         {
             sourceTokensCount++;
             auto word = stream->getAttribute<TermAttribute>()->term();
+            sourceTokenPosition += stream->getAttribute<PositionIncrementAttribute>()->getPositionIncrement();
             auto term = newLucene<Term>(sourceField, word);
             boolQ->add(newLucene<TermQuery>(term), BooleanClause::SHOULD);
-            phraseQ->add(term);
+            phraseQ->add(term, sourceTokenPosition);
         }
 
-        auto searcher = newLucene<IndexSearcher>(Reader());
+        auto searcher = m_mng->Searcher();
 
         // Try exact phrase first:
-        PerformSearch(searcher, langQ, source, phraseQ, results,
+        PerformSearch(searcher.ptr(), srclangQ, langQ, source, phraseQ, results,
                       QUALITY_THRESHOLD, /*scoreScaling=*/1.0);
         if (!results.empty())
             return results;
 
         // Then, if no matches were found, permit being a bit sloppy:
         phraseQ->setSlop(1);
-        PerformSearch(searcher, langQ, source, phraseQ, results,
+        PerformSearch(searcher.ptr(), srclangQ, langQ, source, phraseQ, results,
                       QUALITY_THRESHOLD, /*scoreScaling=*/0.9);
 
         if (!results.empty())
@@ -308,12 +432,12 @@ SuggestionsList TranslationMemoryImpl::Search(const Language& lang,
         boolQ->setMinimumNumberShouldMatch(std::max(1, boolQ->getClauses().size() - MAX_ALLOWED_LENGTH_DIFFERENCE));
         PerformSearchWithBlock
         (
-            searcher, langQ, source, boolQ,
+            searcher.ptr(), srclangQ, langQ, source, boolQ,
             QUALITY_THRESHOLD, /*scoreScaling=*/0.8,
             [=,&results](DocumentPtr doc, double score)
             {
-                auto s = doc->get(sourceField);
-                auto t = doc->get(L"trans");
+                auto s = get_text_field(doc, sourceField);
+                auto t = get_text_field(doc, L"trans");
                 auto stream2 = m_analyzer->tokenStream(sourceField, newLucene<StringReader>(s));
                 int tokensCount2 = 0;
                 while (stream2->incrementToken())
@@ -343,7 +467,7 @@ void TranslationMemoryImpl::GetStats(long& numDocs, long& fileSize)
 {
     try
     {
-        auto reader = Reader();
+        auto reader = m_mng->Reader();
         numDocs = reader->numDocs();
         fileSize = wxDir::GetTotalSize(GetDatabaseDir()).GetValue();
     }
@@ -357,33 +481,32 @@ void TranslationMemoryImpl::GetStats(long& numDocs, long& fileSize)
 class TranslationMemoryWriterImpl : public TranslationMemory::Writer
 {
 public:
-    TranslationMemoryWriterImpl(DirectoryPtr dir, AnalyzerPtr analyzer)
-    {
-        m_writer = newLucene<IndexWriter>(dir, analyzer, IndexWriter::MaxFieldLengthLIMITED);
-        m_writer->setMergeScheduler(newLucene<SerialMergeScheduler>());
-    }
+    TranslationMemoryWriterImpl(IndexWriterPtr writer) : m_writer(writer) {}
 
-    ~TranslationMemoryWriterImpl()
-    {
-        if (m_writer)
-            m_writer->rollback();
-    }
+    ~TranslationMemoryWriterImpl() {}
 
-    virtual void Commit()
+    void Commit() override
     {
         try
         {
-            m_writer->close();
-            m_writer.reset();
+            m_writer->commit();
         }
         CATCH_AND_RETHROW_EXCEPTION
     }
 
-    virtual void Insert(const Language& lang,
-                        const std::wstring& source,
-                        const std::wstring& trans)
+    void Rollback() override
     {
-        if (!lang.IsValid())
+        try
+        {
+            m_writer->rollback();
+        }
+        CATCH_AND_RETHROW_EXCEPTION
+    }
+
+    void Insert(const Language& srclang, const Language& lang,
+                const std::wstring& source, const std::wstring& trans) override
+    {
+        if (!lang.IsValid() || !srclang.IsValid() || lang == srclang)
             return;
 
         // Compute unique ID for the translation:
@@ -392,7 +515,7 @@ public:
           boost::uuids::string_generator()("6e3f73c5-333f-4171-9d43-954c372a8a02");
         boost::uuids::name_generator gen(s_namespace);
 
-        std::wstring itemId(L"en"); // TODO: srclang
+        std::wstring itemId(srclang.WCode());
         itemId += lang.WCode();
         itemId += source;
         itemId += trans;
@@ -406,9 +529,11 @@ public:
 
             doc->add(newLucene<Field>(L"uuid", itemUUID,
                                       Field::STORE_YES, Field::INDEX_NOT_ANALYZED));
+            doc->add(newLucene<Field>(L"v", L"1",
+                                      Field::STORE_YES, Field::INDEX_NO));
             doc->add(newLucene<Field>(L"created", DateField::timeToString(time(NULL)),
                                       Field::STORE_YES, Field::INDEX_NO));
-            doc->add(newLucene<Field>(L"srclang", L"en",
+            doc->add(newLucene<Field>(L"srclang", srclang.WCode(),
                                       Field::STORE_YES, Field::INDEX_NOT_ANALYZED));
             doc->add(newLucene<Field>(L"lang", lang.WCode(),
                                       Field::STORE_YES, Field::INDEX_NOT_ANALYZED));
@@ -422,38 +547,43 @@ public:
         CATCH_AND_RETHROW_EXCEPTION
     }
 
-    virtual void Insert(const Catalog &cat)
+    void Insert(const Language& srclang, const Language& lang, const CatalogItemPtr& item) override
     {
-        auto lang = cat.GetLanguage();
-        if (!lang.IsValid())
+        if (!lang.IsValid() || !srclang.IsValid())
             return;
 
-        int cnt = cat.GetCount();
-        for (int i = 0; i < cnt; i++)
+        // ignore translations with errors in them
+        if (item->GetValidity() == CatalogItem::Val_Invalid)
+            return;
+
+        // can't handle plurals yet (TODO?)
+        if (item->HasPlural())
+            return;
+
+        // ignore untranslated or unfinished translations
+        if (item->IsFuzzy() || !item->IsTranslated())
+            return;
+
+        Insert(srclang, lang, item->GetString().ToStdWstring(), item->GetTranslation().ToStdWstring());
+    }
+
+    void Insert(const CatalogPtr& cat) override
+    {
+        auto srclang = cat->GetSourceLanguage();
+        auto lang = cat->GetLanguage();
+        if (!lang.IsValid() || !srclang.IsValid())
+            return;
+
+        for (auto& item: cat->items())
         {
-            const CatalogItem& item = cat[i];
-
-            // ignore translations with errors in them
-            if (item.GetValidity() == CatalogItem::Val_Invalid)
-                continue;
-
-            // can't handle plurals yet (TODO?)
-            if (item.HasPlural())
-                continue;
-
-            // ignore untranslated or unfinished translations
-            if (item.IsFuzzy() || !item.IsTranslated())
-                continue;
-
             // Note that dt.IsModified() is intentionally not checked - we
             // want to save old entries in the TM too, so that we harvest as
             // much useful translations as we can.
-
-            Insert(lang, item.GetString().ToStdWstring(), item.GetTranslation().ToStdWstring());
+            Insert(srclang, lang, item);
         }
     }
 
-    virtual void DeleteAll()
+    void DeleteAll() override
     {
         try
         {
@@ -466,11 +596,21 @@ private:
     IndexWriterPtr m_writer;
 };
 
-std::shared_ptr<TranslationMemory::Writer> TranslationMemoryImpl::CreateWriter()
+
+void TranslationMemoryImpl::Init()
 {
     try
     {
-        return std::make_shared<TranslationMemoryWriterImpl>(newLucene<DirectoryType>(m_dir), m_analyzer);
+        auto dir = newLucene<DirectoryType>(GetDatabaseDir());
+        m_analyzer = newLucene<StandardAnalyzer>(LuceneVersion::LUCENE_CURRENT);
+
+        m_writer = newLucene<IndexWriter>(dir, m_analyzer, IndexWriter::MaxFieldLengthLIMITED);
+        m_writer->setMergeScheduler(newLucene<SerialMergeScheduler>());
+
+        // get the associated realtime reader & searcher:
+        m_mng.reset(new SearcherManager(m_writer));
+
+        m_writerAPI = std::make_shared<TranslationMemoryWriterImpl>(m_writer);
     }
     CATCH_AND_RETHROW_EXCEPTION
 }
@@ -487,11 +627,7 @@ TranslationMemory *TranslationMemory::ms_instance = nullptr;
 TranslationMemory& TranslationMemory::Get()
 {
     std::call_once(initializationFlag, []() {
-        try
-        {
-            ms_instance = new TranslationMemory;
-        }
-        CATCH_AND_RETHROW_EXCEPTION
+        ms_instance = new TranslationMemory;
     });
     return *ms_instance;
 }
@@ -505,7 +641,18 @@ void TranslationMemory::CleanUp()
     }
 }
 
-TranslationMemory::TranslationMemory() : m_impl(new TranslationMemoryImpl) {}
+TranslationMemory::TranslationMemory() : m_impl(nullptr)
+{
+    try
+    {
+        m_impl = new TranslationMemoryImpl;
+    }
+    catch (...)
+    {
+        m_error = std::current_exception();
+    }
+}
+
 TranslationMemory::~TranslationMemory() { delete m_impl; }
 
 
@@ -513,20 +660,24 @@ TranslationMemory::~TranslationMemory() { delete m_impl; }
 // public API
 // ----------------------------------------------------------------
 
-SuggestionsList TranslationMemory::Search(const Language& lang,
+SuggestionsList TranslationMemory::Search(const Language& srclang,
+                                          const Language& lang,
                                           const std::wstring& source)
 {
-    return m_impl->Search(lang, source);
+    if (!m_impl)
+        std::rethrow_exception(m_error);
+    return m_impl->Search(srclang, lang, source);
 }
 
-void TranslationMemory::SuggestTranslation(const Language& lang,
+void TranslationMemory::SuggestTranslation(const Language& srclang,
+                                           const Language& lang,
                                            const std::wstring& source,
                                            success_func_type onSuccess,
                                            error_func_type onError)
 {
     try
     {
-        onSuccess(Search(lang, source));
+        onSuccess(Search(srclang, lang, source));
     }
     catch (...)
     {
@@ -534,12 +685,16 @@ void TranslationMemory::SuggestTranslation(const Language& lang,
     }
 }
 
-std::shared_ptr<TranslationMemory::Writer> TranslationMemory::CreateWriter()
+std::shared_ptr<TranslationMemory::Writer> TranslationMemory::GetWriter()
 {
-    return m_impl->CreateWriter();
+    if (!m_impl)
+        std::rethrow_exception(m_error);
+    return m_impl->GetWriter();
 }
 
 void TranslationMemory::GetStats(long& numDocs, long& fileSize)
 {
+    if (!m_impl)
+        std::rethrow_exception(m_error);
     m_impl->GetStats(numDocs, fileSize);
 }
